// -------------------------------------------------------------- -*- C++ -*-
// File: sched_tasks.mod
// --------------------------------------------------------------------------
/*******************************************************************************
* Licensed Materials - Property of IBM                                         * 
* 5724-Y55                                                                     *
* (c) Copyright IBM Corporation 1990,2009. All Rights Reserved.                *
* --------------------------------------------------------------------------   *
*                                                                              *
* Problem Description : model of staff scheduling.                             *
*                                                                              *
* A resource is associated with a ResourceType as a capability                 *
* to execute a task.                                                           *
*                                                                              *
* A request is associated with a RequestType as the product of its             *
* execution and a due date.                                                    *
*                                                                              *
* A task is a TaskType and a processing time.                                  *
*                                                                              *
* The recipies associate a RequestType, the tasks to perform, and the          *
* temporal dependencies between tasks (precedence).                            *
*                                                                              *
* A requirement is a TaskType, a ResourceType and a quantity representing      *
* the number of resources to be used by a task.                                *
*                                                                              *
* An operation is a task and a resource deduced from the elements              *
* of the requirement.                                                          * 
********************************************************************************/


// Model and Constraints
//
// Each request, task, and operation is an interval variable. 
// A request spans its tasks.  
// Operations are optional and synchronized with their tasks.
// Each resource is a sequence of non-overlaping intervals.
// The sum of execution of operations for a task and a ResourceType
// is equal to the quantity of the requirement. 
//
// Redundant cumul.
//
// A ResourceType is compatible with several resources and a task 
// requires a defined quantity of ResourceType. A cumul function per ResourceType 
// is defined for the tasks whose requirements needs the ResourceType 
// for the requirement quantity.
// The cumul is limited by the number of resources of the ResourceType.
// These redundant cumuls are crucial as they enforce a stronger constraint 
// when the whole set of resources for the tasks is not chosen. 
// For more complex problems, e.g resources with several ResourceType, 
// other partitions of the resources define efficient redundant cumuls. 

using CP;

// Data source for resources, requests and tasks

{string} ResourceTypes = ...;
{string} RequestTypes  = ...;
{string} TaskTypes     = ...;

tuple RequestDat {
    key int id;
    string  type;
    int     duedate;
    string  name;
};

tuple ResourceDat {
    key int id;
    string  type;
    string name;
};

tuple TaskDat {
    key int id;
    string  type;
    int     ptime;
    string  name;
};

{RequestDat}  requests  = ...;
{ResourceDat} resources = ...;
{TaskDat}     tasks     = ...;

// Data source for template recipes, dependencies and requirements

tuple Recipe {
    string request;
    string task;
};

tuple Dependency {
    string request;
    string taskb;
    string taska;
    int    delay;
};

tuple Requirement {
    string task;
    string resource;
    int    quantity;
};

{Recipe}      recipes      = ...;
{Dependency}  dependencies = ...;
{Requirement} requirements = ...;

// set of demands (task of a request) and operations (task using a resource)

tuple Demand {
    RequestDat request;
    TaskDat    task;
};

{Demand} demands = {<r, t> | r in requests,  m in recipes, t in tasks : r.type == m.request && t.type == m.task};

dvar interval titasks[d in demands] size d.task.ptime;
dvar interval tirequests[requests];

minimize sum (t in requests) maxl(0, endOf(tirequests[t]) - t.duedate);
subject to {
   forall(d1, d2 in demands, dep in dependencies : d1.request == d2.request && 
                                                   dep.taskb == d1.task.type &&
                                                   dep.taska == d2.task.type)
      endBeforeStart(titasks[d1], titasks[d2], dep.delay);
      
    forall(r in requests) 
      span(tirequests[r], all(d in demands : d.request == r) titasks[d]);        


};
